"""Class for metadata maniplation"""

from mscxyz.score_file_classes import MscoreXmlTree
from mscxyz.utils import get_args, color
import json
import lxml
import re
import tmep
import typing


class ReadOnlyFieldError(Exception):

    def __init__(self, field: str):
        self.msg = 'The field “{}” is read only!'.format(field)
        Exception.__init__(self, self.msg)


class UnkownFieldError(Exception):

    def __init__(self, field: str, valid_fields: typing.Sequence):
        self.msg = 'Unkown field of name “{}”! Valid field names are: {}' \
              .format(field, ', '.join(valid_fields))
        Exception.__init__(self, self.msg)


class UnmatchedFormatStringError(Exception):

    def __init__(self, format_string: str, input_string: str):
        self.msg = 'Your format string “{}” doesn’t match on this ' \
              'input string: “{}”'.format(format_string, input_string)
        Exception.__init__(self, self.msg)


class FormatStringNoFieldError(Exception):

    def __init__(self, format_string: str):
        self.msg = 'No fields found in your format string “{}”!' \
              .format(format_string)
        Exception.__init__(self, self.msg)


def distribute_field(source, format_string: str):
    fields = re.findall(r'\$([a-z_]*)', format_string)
    if not fields:
        raise FormatStringNoFieldError(format_string)
    regex = re.sub(r'\$[a-z_]*', '(.*)', format_string)
    match = re.search(regex, source)
    if not match:
        raise UnmatchedFormatStringError(format_string, source)
    values = match.groups()
    return dict(zip(fields, values))


def to_underscore(field):
    return re.sub('([A-Z]+)', r'_\1', field).lower()


def export_to_dict(obj, fields):
    out = {}
    for field in fields:
        value = getattr(obj, field)
        if not value:
            value = ''
        out[field] = value
    return out


class MetaTag(object):

    """The available metaTag fields are:

        * `arranger`
        * `composer`
        * `copyright`
        * `creationDate`
        * `lyricist`
        * `movementNumber`
        * `movementTitle`
        * `platform`
        * `poet`
        * `source`
        * `translator`
        * `workNumber`
        * `workTitle`

        """

    fields = (
        'arranger',
        'composer',
        'copyright',
        'creationDate',
        'lyricist',
        'movementNumber',
        'movementTitle',
        'platform',
        'poet',
        'source',
        'translator',
        'workNumber',
        'workTitle',
    )

    @staticmethod
    def _to_camel_case(field):
        return re.sub(r'(?!^)_([a-zA-Z])',
                      lambda match: match.group(1).upper(), field)

    def __init__(self, xml_root):
        self.xml_root = xml_root

    def _get_element(self, field: str):
        for element in self.xml_root.xpath('//metaTag[@name="' + field + '"]'):
            return element

    def _get_text(self, field: str) -> str:
        element = self._get_element(field)
        if hasattr(element, 'text'):
            return element.text

    def __getattr__(self, field):
        field = self._to_camel_case(field)
        if field not in self.fields:
            raise UnkownFieldError(field, self.fields)
        else:
            return self._get_text(field)

    def __setattr__(self, field, value):
        if field == 'xml_root' or field == 'fields':
            self.__dict__[field] = value
        else:
            field = self._to_camel_case(field)
            self._get_element(field).text = value

    def clean(self):
        fields = (
            'arranger',
            'copyright',
            'creationDate',
            'movementNumber',
            'platform',
            'poet',
            'source',
            'translator',
            'workNumber',
        )
        for field in fields:
            setattr(self, field, '')


class Vbox(object):
    """The first vertical box of a score.

    Available fields:

    * `Composer`
    * `Lyricist`
    * `Subtitle`
    * `Title`

    .. code-block:: xml

        <Staff id="1">
          <VBox>
            <height>10</height>
            <Text>
              <style>Title</style>
              <text>Title</text>
              </Text>
            <Text>
              <style>Composer</style>
              <text>Composer</text>
              </Text>
            </VBox>
          </Staff>

    """

    fields = (
        'Composer',
        'Lyricist',
        'Subtitle',
        'Title',
    )

    def __init__(self, xml_root):
        self.xml_root = xml_root
        xpath = '/museScore/Score/Staff[@id="1"]'
        if not xml_root.xpath(xpath + '/VBox'):
            vbox = lxml.etree.Element('VBox')
            height = lxml.etree.SubElement(vbox, 'height')
            height.text = '10'

            for element in xml_root.xpath(xpath):
                element.insert(0, vbox)

    def _get_tag(self, style):
        """
        :param string style: String inside the `<style>` tags
        """
        for element in self.xml_root.xpath('//VBox/Text'):
            if element.find('style').text == style:
                return element.find('text')

    def _get_text(self, style):
        """
        :param string style: String inside the `<style>` tags
        """
        element = self._get_tag(style)
        if hasattr(element, 'text'):
            return element.text

    def __getattr__(self, field):
        field = field.title()
        if field not in self.fields:
            raise UnkownFieldError(field, self.fields)
        else:
            return self._get_text(field)

    def _create_text_tag(self, style, text):
        """
        :param string style: String inside the `<style>` tags
        """
        Text_tag = lxml.etree.Element('Text')
        style_tag = lxml.etree.SubElement(Text_tag, 'style')
        style_tag.text = style
        text_tag = lxml.etree.SubElement(Text_tag, 'text')
        text_tag.text = text
        for element in self.xml_root.xpath('//VBox'):
            element.append(Text_tag)

    def _set_text(self, style, text):
        """
        :param string style: String inside the `<style>` tags
        """
        element = self._get_tag(style)
        if hasattr(element, 'text'):
            element.text = text
        else:
            self._create_text_tag(style, text)

    def __setattr__(self, field, value):
        if field == 'xml_root' or field == 'fields':
            self.__dict__[field] = value
        elif field.title() not in self.fields:
            raise UnkownFieldError(field, self.fields)
        else:
            self._set_text(field.title(), value)


class Combined(MscoreXmlTree):

    fields = (
        'composer',
        'lyricist',
        'subtitle',
        'title',
    )

    def __init__(self, xml_root):
        self.xml_root = xml_root
        self.metatag = MetaTag(xml_root)
        self.vbox = Vbox(xml_root)

    def _pick_value(self, *values):
        for value in values:
            if value:
                return value

    @property
    def title(self):
        return self._pick_value(self.vbox.Title, self.metatag.workTitle)

    @title.setter
    def title(self, value):
        self.vbox.Title = self.metatag.workTitle = value

    @property
    def subtitle(self):
        return self._pick_value(self.vbox.Subtitle, self.metatag.movementTitle)

    @subtitle.setter
    def subtitle(self, value):
        self.vbox.Subtitle = self.metatag.movementTitle = value

    @property
    def composer(self):
        return self._pick_value(self.vbox.Composer, self.metatag.composer)

    @composer.setter
    def composer(self, value):
        self.vbox.Composer = self.metatag.composer = value

    @property
    def lyricist(self):
        return self._pick_value(self.vbox.Lyricist, self.metatag.lyricist)

    @lyricist.setter
    def lyricist(self, value):
        self.vbox.Lyricist = self.metatag.lyricist = value


class InterfaceReadWrite(object):

    objects = ('metatag', 'vbox', 'combined')

    def __init__(self, xml_root):
        self.metatag = MetaTag(xml_root)
        self.vbox = Vbox(xml_root)
        self.combined = Combined(xml_root)
        self.fields = self.get_all_fields()

    @staticmethod
    def get_all_fields():
        fields = []
        for field in MetaTag.fields:
            fields.append('metatag_' + to_underscore(field))
        for field in Vbox.fields:
            fields.append('vbox_' + field.lower())
        for field in Combined.fields:
            fields.append('combined_' + field)
        return sorted(fields)

    @staticmethod
    def _split(field):
        match = re.search(r'([^_]*)_(.*)', field)
        if not match:
            raise ValueError('Field “' + field + '” can’t be splitted!')
        matches = match.groups()

        if not matches[0] in InterfaceReadWrite.objects:
            raise ValueError(matches[0] + ': Not a supported object!')
        return {'object': matches[0], 'field': matches[1]}

    def export_to_dict(self):
        return export_to_dict(self, self.fields)

    def __getattr__(self, field):
        parts = self._split(field)
        obj = getattr(self, parts['object'])
        return getattr(obj, parts['field'])

    def __setattr__(self, field, value):
        if field in ('fields', 'metatag', 'objects', 'vbox', 'combined'):
            self.__dict__[field] = value
        else:
            parts = self._split(field)
            obj = getattr(self, parts['object'])
            return setattr(obj, parts['field'], value)


class InterfaceReadOnly(object):

    fields = [
        'readonly_abspath',
        'readonly_basename',
        'readonly_dirname',
        'readonly_extension',
        'readonly_filename',
        'readonly_relpath',
        'readonly_relpath_backup',
    ]

    def __init__(self, tree):
        self.xml_tree = tree

    @property
    def readonly_abspath(self):
        return self.xml_tree.abspath

    @property
    def readonly_basename(self):
        return self.xml_tree.basename

    @property
    def readonly_dirname(self):
        return self.xml_tree.dirname

    @property
    def readonly_extension(self):
        return self.xml_tree.extension

    @property
    def readonly_filename(self):
        return self.xml_tree.filename

    @property
    def readonly_relpath(self):
        return self.xml_tree.relpath

    @property
    def readonly_relpath_backup(self):
        return self.xml_tree.relpath_backup


class Interface(object):

    def __init__(self, tree):
        self.xml_tree = tree
        self.read_only = InterfaceReadOnly(tree)
        self.read_write = InterfaceReadWrite(tree.xml_root)
        self.fields = self.get_all_fields()

    @staticmethod
    def get_all_fields():
        return sorted(InterfaceReadOnly.fields +
                      InterfaceReadWrite.get_all_fields())

    def export_to_dict(self):
        return export_to_dict(self, self.fields)

    def __getattr__(self, field):
        if re.match(r'^readonly_', field):
            return getattr(self.read_only, field)
        else:
            return getattr(self.read_write, field)

    def __setattr__(self, field, value):
        if field in ('xml_tree', 'read_only', 'read_write', 'fields'):
            self.__dict__[field] = value
        elif not re.match(r'^readonly_', field):
            return setattr(self.read_write, field, value)
        else:
            raise ReadOnlyFieldError(field)


class Meta(MscoreXmlTree):

    def __init__(self, relpath):
        super(Meta, self).__init__(relpath)

        if not self.errors:
            self.metatag = MetaTag(self.xml_root)
            self.vbox = Vbox(self.xml_root)
            self.combined = Combined(self.xml_root)
            self.interface_read_write = InterfaceReadWrite(self.xml_root)
            self.interface = Interface(self)

    def sync_fields(self):
        if not self.errors:
            self.combined.title = self.combined.title
            self.combined.subtitle = self.combined.subtitle
            self.combined.composer = self.combined.composer
            self.combined.lyricist = self.combined.lyricist

    def distribute_field(self, source_fields, format_string):
        source_fields = source_fields.split(',')
        for source_field in source_fields:
            try:
                source = getattr(self.interface, source_field)
                results = distribute_field(source, format_string)
                if results:
                    for field, value in results.items():
                        setattr(self.interface, field, value)
                return
            except UnmatchedFormatStringError as error:
                self.errors.append(error)

    def write_to_log_file(self, log_file, format_string):
        log = open(log_file, 'w')
        log.write(tmep.parse(format_string, self.interface.export_to_dict()) +
                  '\n')
        log.close()

    def set_field(self, destination_field, format_string):
        field_value = tmep.parse(format_string,
                                 self.interface.export_to_dict())
        setattr(self.interface, destination_field, field_value)

    def clean(self, fields):
        fields = fields[0]
        if fields == 'all':
            fields = self.interface_read_write.fields
        else:
            fields = fields.split(',')
        for field in fields:
            setattr(self.interface_read_write, field, '')

    def delete_duplicates(self):
        iface = self.interface
        if iface.combined_lyricist == iface.combined_composer:
            iface.combined_lyricist = ''

        if not iface.combined_title and iface.combined_subtitle:
            iface.combined_title = iface.combined_subtitle

        if iface.combined_subtitle == iface.combined_title:
            iface.combined_subtitle = ''

    def show(self, pre, post):
        args = get_args()

        fields = list(self.interface.fields)

        if args.general_verbose < 1:
            fields.remove('readonly_abspath')
            fields.remove('readonly_dirname')
            fields.remove('readonly_extension')
            fields.remove('readonly_filename')
            fields.remove('readonly_relpath')

        if args.general_verbose < 2:
            fields.remove('readonly_relpath_backup')

        for field in fields:
            if (args.general_verbose == 0 and (pre[field] or post[field])) or \
               args.general_verbose > 0:

                if re.match(r'^combined_',  field):
                    field_color = 'green'
                elif re.match(r'^metatag_',  field):
                    field_color = 'blue'
                elif re.match(r'^readonly_',  field):
                    field_color = 'red'
                elif re.match(r'^vbox_',  field):
                    field_color = 'cyan'
                else:
                    field_color = 'white'

                line = []
                if pre[field]:
                    line.append('“{}”'.format(pre[field]))

                if pre[field] != post[field]:
                    line.append('->')
                    line.append(color('“{}”'.format(post[field]), 'yellow'))

                print('{}: {}'.format(color(field, field_color),
                                      ' '.join(line)))

    def export_json(self):
        data = {}
        data['title'] = self.get('title')

        output = open(self.relpath.replace(
            '.' + self.extension, '.json'), 'w')
        json.dump(data, output, indent=4)
        output.close()



"""Rename MuseScore files"""

from mscxyz.meta import Meta
from mscxyz.utils import color, get_args
from tmep.format import alphanum, asciify, nowhitespace
import errno
import hashlib
import os
import tmep
import shutil


def create_dir(path: str):
    try:
        os.makedirs(os.path.dirname(path))
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def prepare_fields(fields: dict) -> dict:
    args = get_args()
    out = {}
    for field, value in fields.items():
        if value:
            if args.rename_alphanum:
                value = alphanum(value)
                value = value.strip()
            if args.rename_ascii:
                value = asciify(value)
            if args.rename_no_whitespace:
                value = nowhitespace(value)
            value = value.strip()
            value = value.replace('/', '-')
        out[field] = value
    return out


def apply_format_string(fields: dict) -> str:
    args = get_args()
    fields = prepare_fields(fields)
    name = tmep.parse(args.rename_format, fields)
    return name


def show(old: str, new: str):
    print('{} -> {}'.format(color(old, 'yellow'), color(new, 'green')))


def get_checksum(filename: str) -> str:
    BLOCKSIZE = 65536
    hasher = hashlib.sha1()
    with open(filename, 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)
    return hasher.hexdigest()


def rename_filename(source: str) -> Meta:
    args = get_args()

    meta = Meta(source)
    meta_values = meta.interface.export_to_dict()
    target_filename = apply_format_string(meta_values)

    if args.rename_skip:
        skips = args.rename_skip.split(',')
        for skip in skips:
            if not meta_values[skip]:
                print(color('Field “{}” is empty! Skipping'.format(skip),
                            'red'))
                return meta

    if args.rename_target:
        target_base = os.path.abspath(args.rename_target)
    else:
        target_base = os.getcwd()

    target = os.path.join(target_base,
                          target_filename + '.' + meta.extension)

    if os.path.exists(target):
        target_format = target.replace('.mscx', '{}.mscx')
        i = ''
        while os.path.exists(target_format.format(i)):
            target = target_format.format(i)
            if get_checksum(source) == get_checksum(target):
                print(color('The file “{}” with the same checksum (sha1) '
                            'already exists in the target path “{}”!'
                            .format(source, target), 'red'))
                return meta
            if i == '':
                i = 1
            i += 1

        target = target_format.format(i)

    show(source, target)

    if not args.general_dry_run:

        create_dir(target)
        # Invalid cross-device link:
        # os.rename(source, target)
        shutil.move(source, target)

    return meta



"""A collection of classes intended to represent one MuseScore file.

The classes build on each other hierarchically. The class hierarchy:

.. code ::

    MscoreFile
        MscoreXmlTree
            MscoreStyleInterface
            MscoreLyricsInterface
"""


from mscxyz.utils import mscore, re_open
import lxml.etree  # Needed for type hints
import fnmatch
import lxml
import os
import shutil
import string
import zipfile
import tempfile


def list_scores(path: str, extension: str = 'both', glob: str = None) -> list:
    """List all scores in path.

    :param path: The path so search for score files.
    :param extension: Possible values: “both”, “mscz” or “mscx”.
    :param glob: A glob string, see fnmatch
    """
    if not glob:
        if extension == 'both':
            glob = '*.msc[xz]'
        elif extension in ('mscx', 'mscz'):
            glob = '*.{}'.format(extension)
        else:
            raise ValueError('Possible values for the argument “extension” '
                             'are: “both”, “mscx”, “mscz”')
    if os.path.isfile(path):
        if fnmatch.fnmatch(path, glob):
            return [path]
        else:
            return []
    out = []
    for root, _, scores in os.walk(path):
        for score in scores:
            if fnmatch.fnmatch(score, glob):
                scores_path = os.path.join(root, score)
                out.append(scores_path)
    out.sort()
    return out


def list_zero_alphabet() -> list:
    """Build a list: 0, a, b, c etc."""
    score_dirs = ['0']
    for char in string.ascii_lowercase:
        score_dirs.append(char)
    return score_dirs


###############################################################################
# Class hierarchy level 1
###############################################################################


class MscoreFile(object):
    """This class holds basic file properties of the MuseScore score file.

    :param relpath: The relative (or absolute) path of a MuseScore
        file.
    """

    def __init__(self, relpath: str):
        self.errors = []
        """A list to store errors."""

        self.relpath = relpath
        """The relative path of the score file, for example:
        ``files_mscore2/simple.mscx``.
        """

        self.abspath = os.path.abspath(relpath)
        """The absolute path of the score file, for example:
        ``/home/jf/test/files_mscore2/simple.mscx``."""

        self.extension = relpath.split('.')[-1].lower()
        """The extension (``mscx`` or ``mscz``) of the score file, for
        example: ``mscx``."""

        self.relpath_backup = relpath.replace(
            '.' + self.extension, '_bak.' + self.extension)
        """The backup path of the score file, for example:
        ``files_mscore2/simple_bak.mscx``."""

        self.dirname = os.path.dirname(relpath)
        """The name of the containing directory of the MuseScore file, for
        example: ``files_mscore2``."""

        self.filename = os.path.basename(relpath)
        """The filename of the MuseScore file, for example:
        ``simple.mscx``."""

        self.basename = self.filename.replace('.mscx', '')
        """The basename of the score file, for example: ``simple``."""

        if self.extension == 'mscz':
            self.loadpath = self._unzip(self.abspath)
            """The load path of the score file"""

        else:
            self.loadpath = self.abspath

    @staticmethod
    def _unzip(abspath: str):
        tmp_zipdir = tempfile.mkdtemp()
        zip_ref = zipfile.ZipFile(abspath, 'r')
        zip_ref.extractall(tmp_zipdir)
        zip_ref.close()
        con = os.path.join(tmp_zipdir, 'META-INF', 'container.xml')
        container_info = lxml.etree.parse(con)
        mscx = container_info \
            .xpath('string(/container/rootfiles/rootfile/@full-path)')
        return os.path.join(tmp_zipdir, mscx)

    def backup(self):
        """Make a copy of the MuseScore file."""
        shutil.copy2(self.relpath, self.relpath_backup)

    def export(self, extension: str = 'pdf'):
        """Export the score to the specifed file type.

        :param extension: The extension (default: pdf)
        """
        score = self.relpath
        mscore(['--export-to',
                score.replace('.' + self.extension, '.' + extension), score])


###############################################################################
# Class hierarchy level 2
###############################################################################


class MscoreXmlTree(MscoreFile):
    """XML tree manipulation

    :param relpath: The relative (or absolute) path of a MuseScore file.
    """

    def __init__(self, relpath: str):
        super(MscoreXmlTree, self).__init__(relpath)
        try:
            self.xml_tree = lxml.etree.parse(self.loadpath)
        except lxml.etree.XMLSyntaxError as e:
            self.errors.append(e)
        else:
            self.xml_root = self.xml_tree.getroot()
            musescore = self.xml_tree.xpath('/museScore')
            version = musescore[0].get('version')
            self.version_major = int(version.split('.')[0])
            """The major MuseScore version, for example 2 or 3"""
            self.version = float(version)
            """The MuseScore version, for example 2.03 or 3.01"""

    def add_sub_element(self, root_tag, tag, text: str):
        tag = lxml.etree.SubElement(root_tag, tag)
        tag.text = text

    def strip_tags(self, *tag_names: str):
        """Delete / strip some tag names."""
        lxml.etree.strip_tags(self.xml_tree, tag_names)

    def remove_tags_by_xpath(self, *xpath_strings: str):
        """Remove tags by xpath strings.

        :param xpath_strings: A xpath string.

        .. code:: Python

            tree.remove_tags_by_xpath(
                '/museScore/Score/Style', '//LayoutBreak', '//StemDirection'
            )

        """
        for xpath_string in xpath_strings:
            for rm in self.xml_tree.xpath(xpath_string):
                rm.getparent().remove(rm)

    def merge_style(self, styles: str):
        """Merge styles into the XML tree.

        :param styles: The path of the style file or a string containing
          the XML style markup.

        ``styles`` may not contain surrounding ``<Style>`` tags. This input is
        valid:

        .. code :: XML

            <TextStyle>
              <halign>center</halign>
              <valign>bottom</valign>
              <xoffset>0</xoffset>
              <yoffset>-1</yoffset>
              <offsetType>spatium</offsetType>
              <name>Form Section</name>
              <family>Alegreya Sans</family>
              <size>12</size>
              <bold>1</bold>
              <italic>1</italic>
              <sizeIsSpatiumDependent>1</sizeIsSpatiumDependent>
              <frameWidthS>0.1</frameWidthS>
              <paddingWidthS>0.2</paddingWidthS>
              <frameRound>0</frameRound>
              <frameColor r="0" g="0" b="0" a="255"/>
              </TextStyle>

        This input is invalid:

        .. code :: XML

            <?xml version="1.0"?>
            <museScore version="2.06">
              <Style>
                <TextStyle>
                  <halign>center</halign>
                  <valign>bottom</valign>
                  <xoffset>0</xoffset>
                  <yoffset>-1</yoffset>
                  <offsetType>spatium</offsetType>
                  <name>Form Section</name>
                  <family>Alegreya Sans</family>
                  <size>12</size>
                  <bold>1</bold>
                  <italic>1</italic>
                  <sizeIsSpatiumDependent>1</sizeIsSpatiumDependent>
                  <frameWidthS>0.1</frameWidthS>
                  <paddingWidthS>0.2</paddingWidthS>
                  <frameRound>0</frameRound>
                  <frameColor r="0" g="0" b="0" a="255"/>
                  </TextStyle>
                </Style>
              </museScore>

        """
        if os.path.exists(styles):
            style = lxml.etree.parse(styles).getroot()
        else:
            # <?xml ... tag without encoding to avoid this error:
            # ValueError: Unicode strings with encoding declaration are
            # not supported. Please use bytes input or XML fragments without
            # declaration.
            pre = '<?xml version="1.0"?><museScore version="2.06"><Style>'
            post = '</Style></museScore>'
            style = lxml.etree.XML(pre + styles + post)

        for score in self.xml_tree.xpath('/museScore/Score'):
            score.insert(0, style[0])

    def clean(self):
        """Remove the style, the layout breaks, the stem directions and the
        ``font``, ``b``, ``i``, ``pos``, ``offset`` tags"""
        self.remove_tags_by_xpath(
            '/museScore/Score/Style', '//LayoutBreak', '//StemDirection')
        self.strip_tags('font', 'b', 'i', 'pos', 'offset')

    def save(self, new_name: str = '', mscore: bool = False):
        """Save the MuseScore file.

        :param new_name: Save the MuseScore file under a new name.
        :param mscore: Save the MuseScore file by opening it with the
          MuseScore executable and save it there.
        """
        if new_name:
            filename = new_name
        else:
            filename = self.relpath
        if not self.errors:
            # To get the same xml tag structure as the original score file
            # has.
            for xpath in ('//LayerTag',
                          '//metaTag',
                          '//font',
                          '//i',
                          '//evenFooterL',
                          '//evenFooterC',
                          '//evenFooterR',
                          '//oddFooterL',
                          '//oddFooterC',
                          '//oddFooterR',
                          '//chord/name',
                          '//chord/render',
                          '//StaffText/text',
                          '//Jump/continueAt',
                          ):

                for tag in self.xml_tree.xpath(xpath):
                    if not tag.text:
                        tag.text = ''

            score = open(filename, 'w')
            score.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            score.write(lxml.etree.tostring(self.xml_root, encoding='UTF-8')
                        .decode('utf-8'))
            score.write('\n')
            score.close()
            if mscore:
                re_open(filename)


###############################################################################
# Class hierarchy level 3
###############################################################################

class MscoreStyleInterface(MscoreXmlTree):
    """
    Interface specialized for the style manipulation.

    :param relpath: The relative (or absolute) path of a MuseScore file.
    """
    def __init__(self, relpath: str):
        super(MscoreStyleInterface, self).__init__(relpath)
        styles = self.xml_tree.xpath('/museScore/Score/Style')
        if styles:
            self.style = styles[0]
            """The ``/museScore/Score/Style`` element object, see
            https://lxml.de/tutorial.html#the-element-class
            """
        else:
            self.style = self._create_parent_style()

    def _create_parent_style(self):
        score = self.xml_tree.xpath('/museScore/Score')
        return lxml.etree.SubElement(score[0], 'Style')

    def _create(self, tag: str) -> lxml.etree.Element:
        """
        :param tag: Nested tags are supported, for example ``TextStyle/halign``
        """
        tags = tag.split('/')
        parent = self.style
        for tag in tags:
            element = parent.find(tag)
            if element is None:
                parent = lxml.etree.SubElement(parent, tag)
            else:
                parent = element
        return parent

    def get_element(self, element_path: str,
                    create: bool = False) -> lxml.etree.Element:
        """
        Get a lxml element which is parent to the ``Style`` tag.

        :param element_path: see
          http://lxml.de/tutorial.html#elementpath
        :param create: Create the element if not present in the parent
          ``Style`` tag.

        Example code:

        .. code:: Python

            # Set attributes on a maybe non-existent style tag.
            # <measureNumberOffset x="0.5" y="-2"/>
            test = MscoreStyleInterface('text.mscx')
            element = test.get_element('measureNumberOffset', create=True)
            element.attrib['x'] = '0.5'
            element.attrib['y'] = '-2'
            test.save()
        """
        element = self.style.find(element_path)
        if element is None and create:
            element = self._create(element_path)
        return element

    def get_value(self, element_path: str) -> str:
        """
        Get the value (text) of a style tag.

        :param element_path: see
          http://lxml.de/tutorial.html#elementpath
        """
        element = self.get_element(element_path)
        return element.text

    def set_attributes(self, element_path: str,
                       attributes: dict) -> lxml.etree.Element:
        """Set attributes on a style child tag.

        :param element_path: see
          http://lxml.de/tutorial.html#elementpath
        """
        element = self.get_element(element_path, create=True)
        for name, value in attributes.items():
            element.attrib[name] = str(value)
        return element

    def set_value(self, element_path: str, value: str):
        """
        :param element_path: see
          http://lxml.de/tutorial.html#elementpath
        """
        element = self.style.find(element_path)
        if element is None:
            element = self._create(element_path)
        element.text = str(value)

    def _get_text_style_element(self, name: str) -> lxml.etree.Element:
        if self.version_major != 2:
            raise ValueError(
                'This operation is only allowed for MuseScore 2 score files'
            )
        xpath = '//TextStyle/name[contains(., "{}")]'.format(name)
        child = self.xml_tree.xpath(xpath)
        if child:
            return child[0].getparent()
        else:
            el_text_style = lxml.etree.SubElement(self.style, 'TextStyle')
            el_name = lxml.etree.SubElement(el_text_style, 'name')
            el_name.text = name
            return el_text_style

    def get_text_style(self, name: str) -> dict:
        """Get text styles. Only MuseScore2!

        :param name: The name of the text style.
        """
        text_style = self._get_text_style_element(name)
        out = {}
        for child in text_style.iterchildren():
            out[child.tag] = child.text
        return out

    def set_text_style(self, name: str, values: dict):
        """Set text styles. Only MuseScore2!

        :param name: The name of the text style.
        :param values: A dictionary. The keys are the tag names, values are
          the text values of the child tags, for example
          ``{size: 14, bold: 1}``.
        """
        text_style = self._get_text_style_element(name)
        for element_name, value in values.items():
            el = text_style.find(element_name)
            if el is None:
                el = lxml.etree.SubElement(text_style, element_name)
            el.text = str(value)



"""This submodule provides default parameters for args. Here is the
``args`` object stored from ``argparse``. It can be accessed by the other
submodules using the function `get_args()`."""


class DefaultArguments(object):

    clean_style = None
    export_extension = None
    general_backup = False
    general_colorize = False
    general_config_file = None
    general_dry_run = False
    general_executable = None
    general_glob = '*.mscx'
    general_mscore = False
    general_verbose = 0
    help_markdown = False
    help_rst = False
    lyrics_extract = 'all'
    lyrics_fix = False
    lyrics_remap = None
    meta_clean = None
    meta_json = False
    meta_set = None
    meta_sync = False
    path = '.'
    rename_alphanum = False
    rename_ascii = False
    rename_format = '$combined_title ($combined_composer)'
    rename_no_whitespace = False
    rename_skip = None
    rename_target = None
    subcommand = None


args = DefaultArguments()



"""Wrapper for the command line interface."""

from mscxyz._version import get_versions
from mscxyz.meta import InterfaceReadWrite, Interface
import argparse
import textwrap
import tmep
import typing


def list_fields(fields: typing.Sequence, prefix: str = '',
                suffix: str = '') -> str:
    out = []
    for field in fields:
        out.append(prefix + '- ' + field + suffix)
    return '\n'.join(out)


parser = argparse.ArgumentParser(description='A command \
    line tool to manipulate the XML based "*.mscX" and "*.mscZ" \
    files of the notation software MuseScore.')

###############################################################################
# Global options
###############################################################################

parser.add_argument(
    '-V',
    '--version',
    action='version',
    version='%(prog)s {version}'.format(
        version=get_versions()['version']
    )
)

parser.add_argument(
    '-b',
    '--backup',
    dest='general_backup',
    action='store_true',
    help='Create a backup file.')


parser.add_argument(
    '-c',
    '--colorize',
    action='store_true',
    dest='general_colorize',
    help='Colorize the command line print statements.')

parser.add_argument(
    '-C',
    '--config-file',
    dest='general_config_file',
    help='Specify a configuration file in the INI format.')

parser.add_argument(
    '-d',
    '--dry-run',
    action='store_true',
    dest='general_dry_run',
    help='Simulate the actions.')

parser.add_argument(
    '-g',
    '--glob',
    dest='general_glob',
    default='*.msc[xz]',
    help='Handle only files which matches against Unix style \
    glob patterns (e. g. "*.mscx", "* - *"). If you omit this \
    option, the standard glob pattern "*.msc[xz]" is used.')

parser.add_argument(
    '-m',
    '--mscore',
    action='store_true',
    dest='general_mscore',
    help='Open and save the XML file in MuseScore after manipulating the XML \
    with lxml to avoid differences in the XML structure.')

parser.add_argument(
    '-e',
    '--executable',
    dest='general_executable',
    help='Path of the musescore executable.')

parser.add_argument(
    '-v',
    '--verbose',
    action='count',
    dest='general_verbose',
    default=0,
    help='Make commands more verbose. You can specifiy \
    multiple arguments (. g.: -vvv) to make the command more \
    verbose.')

###############################################################################
# subparser
###############################################################################

subparser = parser.add_subparsers(
    title='Subcommands',
    dest='subcommand',
    help='Run "subcommand --help" for more \
    informations.')

###############################################################################
# clean
###############################################################################

sub_clean = subparser.add_parser(
    'clean', help='Clean and reset the formating of the "*.mscx" file')

sub_clean.add_argument(
    '-s',
    '--style',
    dest='clean_style',
    type=open,
    help='Load a "*.mss" style file and include the contents of \
    this file.')

###############################################################################
# meta
###############################################################################

sub_meta = subparser.add_parser(
    'meta',
    help='Deal with meta data informations stored in the MuseScore file.',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent('''\
    MuseScore can store meta data informations in different places:

    # metatag

    ## XML structure of a meta tag:

        <metaTag name="tag"></metaTag>

    ## All meta tags:

        - arranger
        - composer
        - copyright
        - creationDate
        - lyricist
        - movementNumber
        - movementTitle
        - platform
        - poet
        - source
        - translator
        - workNumber
        - workTitle

    # vbox

    ## XML structure of a vbox tag:

        <VBox>
          <Text>
            <style>Title</style>
            <text>Some title text</text>
            </Text>

    ## All vbox tags:

        - Title
        - Subtitle
        - Composer
        - Lyricist

    This command line tool bundles some meta data informations:

    # Combined meta data fields:

        - title (1. vbox_title 2. metatag_work_title)
        - subtitle (1. vbox_subtitle 2. metatag_movement_title)
        - composer (1. vbox_composer 2. metatag_composer)
        - lyricist (1. vbox_lyricist 2. metatag_lyricist)

    You have access to all this metadata fields through following fields:''')
    + '\n\n' + list_fields(InterfaceReadWrite.get_all_fields(), prefix='    '))

sub_meta.add_argument(
    '-c',
    '--clean',
    nargs=1,
    dest='meta_clean',
    help='Clean the meta data fields. Possible values: „all“ or \
    „field_one,field_two“.')

sub_meta.add_argument(
    '-D',
    '--delete-duplicates',
    dest='meta_delete',
    action='store_true',
    help='Deletes combined_lyricist if this field is equal to '
    'combined_composer. Deletes combined_subtitle if this field is equal to'
    'combined_title. Move combined_subtitle to combimed_title if '
    'combined_title is empty.')

sub_meta.add_argument(
    '-d',
    '--distribute-fields',
    dest='meta_dist',
    action='append',
    nargs=2,
    metavar=('SOURCE_FIELDS', 'FORMAT_STRING'),
    help='Distribute source fields to target fields applying a format string \
    on the source fields. It is possible to apply multiple \
    --distribute-fields options. SOURCE_FIELDS can be a single field or a \
    comma separated list of fields: field_one,field_two. The program \
    tries first to match the FORMAT_STRING on the first source field. If this\
    fails, it tries the second source field ... an so on.')

sub_meta.add_argument(
    '-j',
    '--json',
    action='store_true',
    dest='meta_json',
    help='Additionally write the meta data to a json file.')

sub_meta.add_argument(
    '-l',
    '--log',
    nargs=2,
    metavar=('DESTINATION', 'FORMAT_STRING'),
    dest='meta_log',
    help='Write one line per file to a text file. e. g. --log '
    '/tmp/mscx-manager.log \'$title $composer\'')

sub_meta.add_argument(
    '-s',
    '--synchronize',
    action='store_true',
    dest='meta_sync',
    help='Synchronize the values of the first vertical frame (vbox) \
    (title, subtitle, composer, lyricist) with the corresponding \
    metadata fields')

sub_meta.add_argument(
    '-S',
    '--set-field',
    nargs=2,
    action='append',
    metavar=('DESTINATION_FIELD', 'FORMAT_STRING'),
    dest='meta_set',
    help='Set value to meta data fields.')

###############################################################################
# lyrics
###############################################################################

sub_lyrics = subparser.add_parser(
    'lyrics',
    help='Extract lyrics. Without any option this subcommand \
    extracts all lyrics verses into separate mscx files. \
    This generated mscx files contain only one verse. The old \
    verse number is appended to the file name, e. g.: \
    score_1.mscx.')

sub_lyrics.add_argument(
    '-e',
    '--extract',
    dest='lyrics_extract',
    default='all',
    help='The lyric verse number to extract or "all".'
)

sub_lyrics.add_argument(
    '-r',
    '--remap',
    dest='lyrics_remap',
    help='Remap lyrics. Example: "--remap 3:2,5:3". This \
    example remaps lyrics verse 3 to verse 2 and verse 5 to 3. \
    Use commas to specify multiple remap pairs. One remap pair \
    is separated by a colon in this form: "old:new": "old" \
    stands for the old verse number. "new" stands for the new \
    verse number.')

sub_lyrics.add_argument(
    '-f',
    '--fix',
    action='store_true',
    dest='lyrics_fix',
    help='Fix lyrics: Convert trailing hyphens ("la- la- la") \
    to a correct hyphenation ("la - la - la")')

###############################################################################
# rename
###############################################################################

sub_rename = subparser.add_parser(
    'rename',
    help='Rename the "*.mscx" files.',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description='Fields and functions you can use in the format '
    'string (-f, --format):\n\n'
    'Fields\n======\n\n{}\n\n'
    'Functions\n=========\n\n{}'
    .format(list_fields(Interface.get_all_fields(), prefix='    '),
            tmep.doc.Doc().get()))

sub_rename.add_argument(
    '-f',
    '--format',
    dest='rename_format',
    default='$combined_title ($combined_composer)',
    help='Format string.')

sub_rename.add_argument(
    '-A',
    '--alphanum',
    dest='rename_alphanum',
    action='store_true',
    help='Use only alphanumeric characters.')

sub_rename.add_argument(
    '-a',
    '--ascii',
    dest='rename_ascii',
    action='store_true',
    help='Use only ASCII characters.')

sub_rename.add_argument(
    '-n',
    '--no-whitespace',
    dest='rename_no_whitespace',
    action='store_true',
    help='Replace all whitespaces with dashes or \
    sometimes underlines.')

sub_rename.add_argument(
    '-s',
    '--skip-if-empty',
    dest='rename_skip',
    metavar='FIELDS',
    help='Skip rename action if FIELDS are empty. Separate FIELDS using \
    commas: combined_composer,combined_title',
)

sub_rename.add_argument(
    '-t',
    '--target',
    dest='rename_target',
    help='Target directory',
)

###############################################################################
# export
###############################################################################

sub_export = subparser.add_parser(
    'export',
    help='Export the scores to PDFs or to the specified \
    extension.')

sub_export.add_argument(
    '-e',
    '--extension',
    dest='export_extension',
    default='pdf',
    help='Extension to export. If this option \
    is omitted, then the default extension is "pdf".')

###############################################################################
# help
###############################################################################

sub_help = subparser.add_parser(
    'help',
    help='Show help. Use “{} help all” to show help \
    messages of all subcommands. Use \
    “{} help <subcommand>” to show only help messages \
    for the given subcommand.'.format(parser.prog, parser.prog))

sub_help.add_argument(
    '-m',
    '--markdown',
    dest='help_markdown',
    action='store_true',
    help='Show help in markdown format. \
    This option enables to generate the README file directly \
    form the command line output.')

sub_help.add_argument(
    '-r',
    '--rst',
    dest='help_rst',
    action='store_true',
    help='Show help in reStructuresText \
    format. This option enables to generate the README file \
    directly form the command line output.')

###############################################################################
# help
###############################################################################

parser.add_argument(
    'path',
    help='Path to a "*.mscx" file \
    or a folder which contains "*.mscx" files. In conjunction \
    with the subcommand "help" this positional parameter \
    accepts the names of all other subcommands or the word \
    "all".')



"""A collection of useful utility functions"""

import os
import platform
import subprocess
import termcolor
from typing import Sequence
from mscxyz.settings import DefaultArguments


def get_args() -> DefaultArguments:
    """Get the ``args`` object (the ``argparse`` object) which is stored in
    the .settings.py submodule for all other submodules.

    :return: the ``argparse`` object
    """
    from mscxyz import settings
    return getattr(settings, 'args')


def set_args(args) -> DefaultArguments:
    """Set the ``args`` object (the ``argparse`` object) which is stored in
    the .settings.py submodule for all other submodules to import.
    """
    from mscxyz import settings
    setattr(settings, 'args', args)
    return args


def get_mscore_bin() -> str:
    """Check the existance of the executable mscore

    :return: Path of the executable.
    """
    args = get_args()
    system = platform.system()
    if args and args.general_executable:
        binary = args.general_executable
    elif system == 'Darwin':
        binary = '/Applications/MuseScore 2.app/Contents/MacOS/mscore'
    else:
        cmd = 'where' if system == 'Windows' else 'which'
        binary = subprocess.check_output([cmd, 'mscore'])
        binary = binary.decode('utf-8')
        binary = binary.replace('\n', '')

    if os.path.exists(binary):
        return binary
    else:
        raise ValueError('mscore binary could not be found.')


def mscore(cli_args: Sequence[str]) -> subprocess.Popen:
    """
    :param cli_args: Command line arguments to call the mscore binary with.
    """
    executable = get_mscore_bin()
    cli_args.insert(0, executable)
    p = subprocess.Popen(cli_args, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    p.wait()
    if p.returncode != 0:
        for line in p.stderr:
            print(line.decode('utf-8'))
        raise ValueError('mscore exits with returncode != 0')
    return p


def re_open(input_file: str):
    """Open and save a MuseScore file with the ``mscore`` binary under the same
    file path.

    :param input_file: The path (relative or absolute) of a MuseScore file.
    """
    mscore(['-o', input_file, input_file])


def convert_mxl(input_file: str):
    """
    Convert a MusicXML file into a MuseScore file.

    :param input_file: The path (relative or absolute) of a MusicXML file.
    """
    output_file = input_file.replace('.mxl', '.mscx')
    mscore(['-o', output_file, input_file])
    os.remove(input_file)


def color(*args) -> str:
    """Wrapper function around ``termcolor.colored()`` to easily turn off and
    on colorized terminal output on the command line.

    :param args: First arg must be an string to print out, second arg a color
      name.

    Example usage:

    .. code:: Python

        color('“{}”'.format(post[field]), 'yellow')
    """
    settings = get_args()
    if settings.general_colorize:
        return termcolor.colored(*args)
    else:
        return args[0]



"""Manipulate the lyrics"""

from mscxyz.score_file_classes import MscoreXmlTree
import lxml.etree as etree
import typing


class MscoreLyricsInterface(MscoreXmlTree):

    def __init__(self, relpath: str):
        super(MscoreLyricsInterface, self).__init__(relpath)
        self.lyrics = self.normalize_lyrics()
        self.max = self.get_max()

    def normalize_lyrics(self):
        """Normalize numbering of verses to natural numbering (1,2,3).

        From

        .. code-block:: xml

                <Lyrics>
                        <text>1. la</text>
                        </Lyrics>
                <Lyrics>
                        <no>1</no>
                        <style>Lyrics Even Lines</style>
                <text>2. li</text>
                </Lyrics>
                <Lyrics>
                        <no>2</no>
                        <text>3. lo</text>
                </Lyrics>

        To

        .. code-block:: python

                [
                        {'number': 1, 'element': lyrics_tag},
                        {'number': 2, 'element': lyrics_tag},
                        {'number': 3, 'element': lyrics_tag},
                ]
        """
        lyrics = []
        for lyric in self.xml_tree.findall('.//Lyrics'):
            safe = {}
            safe['element'] = lyric
            number = lyric.find('no')

            if hasattr(number, 'text'):
                no = int(number.text) + 1
            else:
                no = 1
            safe['number'] = no

            lyrics.append(safe)

        return lyrics

    def get_max(self):
        """Retrieve the number of verses.

        From:

                1. La
                2. La
                3. La

        To:

                3

        """
        max_lyric = 0
        for element in self.lyrics:
            if element['number'] > max_lyric:
                max_lyric = element['number']

        return max_lyric

    def remap(self, remap_string: str, mscore: bool = False):
        for pair in remap_string.split(','):
            old = pair.split(':')[0]
            new = pair.split(':')[1]
            for element in self.lyrics:
                if element['number'] == int(old):
                    element['element'].find('no').text = str(int(new) - 1)

        self.save(mscore)

    def extract_one_lyrics_verse(self, number: int, mscore: bool = False):
        """Extract a lyric verse by verse number.

        :param number: The number of the lyrics verse starting by 1
        """
        score = MscoreLyricsInterface(self.relpath)

        for element in score.lyrics:
            tag = element['element']

            if element['number'] != number:
                tag.getparent().remove(tag)
            elif number != 1:
                tag.find('no').text = '0'

        ext = '.' + self.extension
        new_name = score.relpath.replace(ext, '_' + str(number) + ext)
        score.save(new_name, mscore)

    def extract_lyrics(self, number: typing.Union[int, str] = None,
                       mscore: bool = False):
        """Extract one lyric verse or all lyric verses.

        :param mixed number: The lyric verse number or 'all'
        """

        if number == 'all':
            for n in range(1, self.max + 1):
                self.extract_one_lyrics_verse(n)
        else:
            self.extract_one_lyrics_verse(int(number))

    def fix_lyrics_verse(self, verse_number: int):
        """
        from:

        .. code-block:: xml

                <Lyrics>
                        <text>la-</text>
                </Lyrics>
                <Lyrics>
                        <syllabic>end</syllabic>
                        <text>la-</text>
                </Lyrics>
                <Lyrics>
                        <text>la.</text>
                </Lyrics>

        to:

        .. code-block:: xml

                <Lyrics>
                        <syllabic>begin</syllabic>
                        <text>la</text>
                </Lyrics>
                <Lyrics>
                        <syllabic>middle</syllabic>
                        <text>la</text>
                </Lyrics>
                <Lyrics>
                        <syllabic>end</syllabic>
                        <text>la.</text>
                </Lyrics>
        """

        syllabic = False
        for element in self.lyrics:
            if element['number'] == verse_number:
                tag = element['element']
                tag_text = tag.find('text')
                text = tag_text.text
                tag_syl = etree.Element('syllabic')
                if text.endswith('-'):
                    tag_text.text = text[:-1]
                    if not syllabic:
                        tag_syl.text = 'begin'
                        syllabic = True
                    else:
                        tag_syl.text = 'middle'
                else:
                    if syllabic:
                        tag_syl.text = 'end'
                        syllabic = False
                    else:
                        tag_syl = False

                if not isinstance(tag_syl, bool):
                    tag.append(tag_syl)

    def fix_lyrics(self, mscore: bool = False):
        for verse_number in range(1, self.max + 1):
            self.fix_lyrics_verse(verse_number)

        self.save(mscore=False)



"""A command line tool to manipulate the XML based mscX and mscZ
files of the notation software MuseScore.

API Interface:

Classes:

.. code ::

    MscoreFile
        MscoreXmlTree
            MscoreLyricsInterface
            MscoreMetaInterface
            MscoreStyleInterface


Functions:

.. code ::

    exec_mscore_binary
    list_scores
"""

from mscxyz import cli
from mscxyz.lyrics import MscoreLyricsInterface
from mscxyz.meta import Meta
from mscxyz.rename import rename_filename
from mscxyz.score_file_classes import MscoreFile, \
                                      MscoreXmlTree, \
                                      MscoreStyleInterface, \
                                      list_scores
from mscxyz.utils import set_args, color, mscore
from mscxyz.settings import DefaultArguments
import lxml
import sys
import os
import configparser
import typing

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

###############################################################################
# API INTERFACE BEGIN
###############################################################################

# Classes

# Level 1

MscoreFile
"""see submodule ``score_file_classes.py``"""

# Level 2

MscoreXmlTree
"""see submodule ``score_file_classes.py``"""

# Level 3

MscoreLyricsInterface
"""see submodule ``lyrics.py``"""

MscoreMetaInterface = Meta
"""see submodule ``meta.py``"""

MscoreStyleInterface
"""see submodule ``score_file_classes.py``"""

# Functions

exec_mscore_binary = mscore
"""see submodule ``utils.py``"""

list_scores
"""see submodule ``score_file_classes.py``"""

###############################################################################
# API INTERFACE END
###############################################################################


def parse_config_ini(relpath: str = None) -> configparser.ConfigParser:
    """Parse the configuration file. The file format is INI. The default
    location is ``/etc/mscxyz.ini``."""
    if not relpath:
        ini_file = os.path.abspath(os.path.join(os.sep, 'etc', 'mscxyz.ini'))
    else:
        ini_file = relpath
    config = configparser.ConfigParser()
    if os.path.exists(ini_file):
        config.read(ini_file)
        return config


def merge_config_into_args(config: configparser.ConfigParser,
                           args: DefaultArguments) -> DefaultArguments:
    for section in config.sections():
        for key, value in config[section].items():
            arg = '{}_{}'.format(section, key)
            if not hasattr(args, arg) or not getattr(args, arg):
                setattr(args, arg, value)

    for arg in ['general_backup', 'general_colorize', 'general_dry_run',
                'general_mscore', 'help_markdown', 'help_rst',
                'lyrics_fix', 'meta_json', 'meta_sync',
                'rename_alphanum', 'rename_ascii',
                'rename_no_whitespace']:
        if hasattr(args, arg):
            value = getattr(args, arg)
            if value == 1 or value == 'true' or value == 'True':
                setattr(args, arg, True)
            else:
                setattr(args, arg, False)

    return args


def heading(args: DefaultArguments, text: str, level: int = 1):
    length = len(text)
    if args.help_markdown:
        print('\n' + ('#' * level) + ' ' + text + '\n')
    elif args.help_rst:
        if level == 1:
            underline = '='
        elif level == 2:
            underline = '-'
        elif level == 3:
            underline = '^'
        elif level == 4:
            underline = '"'
        else:
            underline = '-'
        print('\n' + text + '\n' + (underline * length) + '\n')
    else:
        print(text)


def code_block(args: DefaultArguments, text: str):
    if args.help_markdown:
        print('```\n' + text + '\n```')
    elif args.help_rst:
        print('.. code-block:: text\n\n  ' + text.replace('\n', '\n  '))
    else:
        print(text)


def show_all_help(args: DefaultArguments):
    subcommands = ('clean', 'meta', 'lyrics', 'rename', 'export', 'help')

    if args.path == 'all':
        heading(args, 'mscxyz', 1)
        code_block(args, cli.parser.format_help())

        heading(args, 'Subcommands', 1)

        for subcommand in subcommands:
            command = getattr(cli, 'sub_' + subcommand)
            heading(args, command.prog, 2)
            code_block(args, command.format_help())

    else:
        code_block(args, getattr(cli, args.path).format_help())


def report_errors(errors: typing.Sequence):
    for error in errors:
        print('{}: {}; message: {}'.format(
            color('Error', 'white', 'on_red'),
            color(error.__class__.__name__, 'red'),
            error.msg
        ))


def no_error(error, errors):
    for e in errors:
        if isinstance(e, error):
            return False
    return True


def execute(args: typing.Sequence = None):
    args = cli.parser.parse_args(args)
    config = parse_config_ini(args.general_config_file)
    if config:
        args = merge_config_into_args(config, args)
    set_args(args)

    if args.subcommand == 'help':
        show_all_help(args)
        sys.exit()

    files = list_scores(path=args.path, glob=args.general_glob)

    for file in files:

        print('\n' + color(file, 'red'))

        if args.general_backup:
            from mscxyz.score_file_classes import MscoreFile
            score = MscoreFile(file)
            score.backup()

        if args.subcommand == 'clean':
            score = MscoreXmlTree(file)
            print(score.filename)
            score.clean()
            if args.clean_style:
                score.merge_style(styles=args.clean_style.name)
            score.save(mscore=args.general_mscore)

        elif args.subcommand == 'lyrics':
            score = MscoreLyricsInterface(file)
            if args.lyrics_remap:
                score.remap(remap_string=args.lyrics_remap,
                            mscore=args.general_mscore)
            elif args.lyrics_fix:
                score.fix_lyrics(mscore=args.general_mscore)
            else:
                score.extract_lyrics(number=args.lyrics_extract,
                                     mscore=args.general_mscore)

        elif args.subcommand == 'meta':
            score = Meta(file)
            if no_error(lxml.etree.XMLSyntaxError, score.errors):
                pre = score.interface.export_to_dict()
                if args.meta_clean:
                    score.clean(fields=args.meta_clean)
                if args.meta_json:
                    score.export_json()
                if args.meta_dist:
                    for a in args.meta_dist:
                        score.distribute_field(source_fields=a[0],
                                               format_string=a[1])
                if args.meta_set:
                    for a in args.meta_set:
                        score.set_field(destination_field=a[0],
                                        format_string=a[1])
                if args.meta_delete:
                    score.delete_duplicates()
                if args.meta_sync:
                    score.sync_fields()
                if args.meta_log:
                    score.write_to_log_file(args.meta_log[0], args.meta_log[1])
                post = score.interface.export_to_dict()
                score.show(pre, post)
            if not args.general_dry_run and not score.errors and pre != post:
                score.save(mscore=args.general_mscore)

        elif args.subcommand == 'rename':
            score = rename_filename(file)

        elif args.subcommand == 'export':
            from mscxyz.score_file_classes import MscoreFile
            score = MscoreFile(file)
            score.export(extension=args.export_extension)

        report_errors(score.errors)


if __name__ == '__main__':
    execute()
